"""Type stubs for classes to be used directly from pynblast"""
from typing import List, Optional, Tuple

class ResamplingArbor:
    def __init__(self, table: List[Tuple[int, Optional[int], float, float, float]]): ...
    def prune_at(self, ids: List[int]) -> int: ...
    def prune_branches_containing(self, ids: List[int]) -> int: ...
    def prune_below_strahler(self, threshold: int) -> int: ...
    def prune_beyond_branches(self, threshold: int) -> int: ...
    def prune_beyond_steps(self, threshold: int) -> int: ...
    def points(self) -> List[Tuple[float, float, float]]: ...
    def skeleton(self) -> List[Tuple[int, Optional[int], float, float, float]]: ...
